#-*- coding: utf-8 -*-
# PackageManager.py
# Wrapper class for package managers.
# Provides their API and allows to use multiple package managers at once.
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

import itertools


class PackageManager:

	"""
	Wrapper class for package managers.
	Provides their API and allows to use multiple package managers at once.

	In actions where it makes no sense to merge results from more package managers,
	result from first manager is used.
	"""

	package_managers = None

	def __init__(self, *instances):
		self.package_managers = instances

	def names(self):
		return map(lambda x: x.__class__.__name__ ,self.package_managers)

	def packages_newer_than(self, unix_time):
		"""
		Returns list of packages which were modified between unix_time and present
		Packages in list should be dictionaries with keys {"name", "modified"}
		"""
		# @FIXME move import to top-level
		from tracer.resources.collections import PackagesCollection
		packages_lists = [p.packages_newer_than(unix_time) for p in self.package_managers]
		return PackagesCollection(itertools.chain.from_iterable(packages_lists))

	def package_files(self, pkg_name):
		"""Returns list of files provided by package"""
		return self.package_managers[0].package_files(pkg_name)

	def load_package_info(self, package):
		"""From database load informations about given package and set them to it"""
		return self.package_managers[0].load_package_info(package)

	def provided_by(self, app):
		"""Returns name of package which provides given application"""
		return self.package_managers[0].provided_by(app)

	def find_package(self, pkg_name, search):
		"""Find a package by name and some other input criteria"""
		return self.package_managers[0].find_package(pkg_name, search)

	def compare_packages(self, package1, package2):
		"""
		Compares two packages by their version information
		Returns:
		0 if they are equal
		1 if package1 > package2
		-1 if package2 > package1
		"""
		return self.package_managers[0].compare_packages(package1, package2)



#-*- coding: utf-8 -*-
# query.py
# Module providing querying operations to Tracer API
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from tracer.resources.tracer import Tracer
from tracer.resources.system import System
from tracer.resources.rules import Rules
from tracer.resources.applications import Applications
from tracer.resources.memory import dump_memory


class Query(object):
	"""
	Provide API for Tracer querying operations.
	They are executed kind of lazily, so running the operation will
	return just an wrapper class with ``get()`` method.

	Example::

		from tracer.query import Query
		q = Query()
		q.affected_applications().get()

	.. note::
		Some querying methods can require root permissions

	"""

	def __init__(self, tracer=Tracer):
		self._tracer = tracer(System.package_manager(), Rules, Applications, dump_memory)

	def from_packages(self, packages):
		"""List of ``Package`` that only should be traced"""
		self._tracer.specified_packages = packages
		return self

	def now(self):
		"""
		Pretend that specified packages have been updated just now.
		Benefit of this is absolutely no need for openning the package history database
		"""
		self._tracer.now = True
		return self

	def affected_applications(self, user=None):
		"""
		Return list of applications which use some outdated files
		"""
		return Lazy(self._tracer.trace_affected, {"user": user})


class Lazy(object):
	def __init__(self, method, kwargs):
		self._method = method
		self._kwargs = kwargs

	def get(self):
		return self._method(**self._kwargs)



from __future__ import print_function
from __future__ import unicode_literals

from sys import __stdout__, version_info


class Arguments(object):
	pass


class View(object):
	args = Arguments()
	out = None

	def __init__(self, out=__stdout__):
		self.out = out

	def assign(self, key, value):
		self.args.__dict__[key] = value

	def get(self, key):
		return self.args.__dict__[key]

	def print(self, text, end=None):
		if version_info.major == 2:
			text = text.encode("utf8")

		print(text, end=end, file=self.out)



#-*- coding: utf-8 -*-
# yum.py
# Module to work with YUM package manager class
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import


from tracer.resources.system import System
if System.distribution() in ["fedora", "rhel", "centos", "centos-7", "ol"]:

	from tracer.packageManagers.rpm import Rpm

	class Yum(Rpm):

		@property
		def history_path(self): return '/var/lib/yum/history/'



from __future__ import unicode_literals

from .. import View
from datetime import datetime


class PackagesView(View):
	def render(self):
		line = "{0: <15}{1: <10}{2: <40}"
		print(line.format("Date", "Time", "Package name"))
		print(55 * "-")
		for package in self.args.packages:
			modified = datetime.fromtimestamp(package.modified)
			date = modified.strftime('%Y-%m-%d')
			time = modified.strftime('%H:%M')
			print(line.format(date, time, package.name))



#-*- coding: utf-8 -*-
# tracer.py
# Tracer finds outdated running applications in your system
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

import os
from psutil import NoSuchProcess
from tracer.resources.package import Package
from tracer.resources.system import System
from tracer.resources.FilenameCleaner import FilenameCleaner
from tracer.resources.processes import AffectedProcess
from tracer.resources.collections import ApplicationsCollection, AffectedProcessesCollection, PackagesCollection
from tracer.resources.exceptions import UnsupportedDistribution
from tracer.resources.applications import Applications, AffectedApplication
from tracer.resources.lang import _


class Tracer(object):
	"""Tracer finds outdated running applications in your system"""

	"""List of packages that only should be traced"""
	specified_packages = None

	"""
	When true, tracer pretends that specified packages have been updated just now.
	Benefit of this is absolutely no need for openning the package history database
	"""
	now = False

	"""
	Timestamp since when the updates should be
	"""
	timestamp = None

	"""Instance of package manager class. Set by __init__"""
	_PACKAGE_MANAGER = None

	"""Structure representing all processes and files they use. See ``tracer.resources.memory.dump_memory``"""
	_memory = None

	"""Objects responsible for providing applications and rules settings form config files"""
	_rules = None
	_applications = None

	"""Observer object responsible for calling desired hooks"""
	_hooks_observer = None

	"""Should tracer find even erased applications?"""
	_erased = False

	def __init__(self, package_manager, rules, applications, memory=None, hooks_observer=None, erased=False):
		if not package_manager:
			raise UnsupportedDistribution(System.distribution())

		self._PACKAGE_MANAGER = package_manager
		self._rules = rules
		self._applications = applications
		self._memory = memory
		self._hooks_observer = hooks_observer
		self._erased = erased

	def _modified_packages(self):
		"""Returns list of packages what tracer should care about"""
		if self.specified_packages and self.now:
			return PackagesCollection(self.specified_packages)

		timestamp = self.timestamp if self.timestamp else System.boot_time()
		packages = self._PACKAGE_MANAGER.packages_newer_than(timestamp)
		packages = packages.intersection(self.specified_packages)
		return packages

	def trace_affected(self, user=None):
		"""
		Returns collection of applications which uses some files that have been modified
		@TODO This function should be hardly optimized
		"""

		memory = self._memory(user)
		packages = self._modified_packages()

		affected = {}
		found = []
		for package in packages.unique_newest():
			for file in self._PACKAGE_MANAGER.package_files(package.name):

				file = FilenameCleaner.strip(file)
				if not file in memory:
					continue

				for p in memory[file]:
					if p.pid in found:
						continue

					try:
						if p.create_time() <= package.modified:
							found.append(p.pid)
							p = self._apply_rules(p)
							a = self._applications.find(p.name())

							if not a.ignore:
								if a.name not in affected:
									if self._erased and not self._PACKAGE_MANAGER.provided_by(a):
										a.type = Applications.TYPES["ERASED"]
									affected[a.name] = AffectedApplication(a._attributes)
									affected[a.name].affected_instances = AffectedProcessesCollection()
									self._call_hook(affected[a.name])
								affected[a.name].affected_instances.append(p)
					except NoSuchProcess:
						pass

		if not self._applications.find('kernel').ignore and self._has_updated_kernel():
			# Add fake AffectedApplication
			affected['kernel'] = AffectedApplication({"name": "kernel", "type": Applications.TYPES["STATIC"],
									"helper": _("You will have to reboot your computer")})

		return ApplicationsCollection(affected.values())

	def _has_updated_kernel(self):
		running = System.running_kernel_package()

		if running is None:
			""" If the running kernel package could not be determined, abort """
			return False

		kernel_package_name = System.kernel_package_name()
		latest = Package(kernel_package_name)
		latest.load_info(self._PACKAGE_MANAGER)

		return self._PACKAGE_MANAGER.compare_packages(running, latest) == -1

	def _apply_rules(self, process):
		parent = process.parent()
		if not parent:
			return process

		c_rule = self._rules.find(process.name())
		p_rule = self._rules.find(parent.name())

		if c_rule and c_rule.action == self._rules.ACTIONS["RETURN"]:
			return process

		if not p_rule or not p_rule.action:
			return process

		if p_rule.action == self._rules.ACTIONS["CALL-PARENT"]:
			return self._apply_rules(parent)

		# Only RETURN action left
		# RETURN rule is defined for parent process
		return parent

	def _call_hook(self, app):
		if self._hooks_observer:
			self._hooks_observer(app.name)

	def trace_application(self, app, affected_process_factory=AffectedProcess):
		"""
		Returns collection of processes where each of them contains
		packages which affected it. Packages contains only files matching
		with the particular process
		"""
		packages = self._modified_packages()
		processes = AffectedProcessesCollection()
		for process in app.instances:
			processes.update(self._affecting_processes(process, packages, affected_process_factory))
			processes.update(self._affecting_children(process, packages, affected_process_factory))
		return processes

	def _affecting_processes(self, process, packages, affected_process_factory=AffectedProcess):
		collection = AffectedProcessesCollection()
		process_files = process.files
		for package in packages:
			matching_files = set()
			for package_file in self._PACKAGE_MANAGER.package_files(package.name):
				package_file = FilenameCleaner.strip(package_file)
				if not package_file in process_files:
					continue

				if process.create_time() <= package.modified:
					matching_files.add(package_file)

			if matching_files:
				aff_pkg = package
				aff_pkg.files = matching_files

				affected = affected_process_factory(process.pid)
				affected.__dict__.update(process.__dict__)
				affected.packages.update([aff_pkg])
				collection.update([affected])
		return collection

	def _affecting_children(self, process, packages, affected_process_factory):
		if not self._rules.find(process.name()):
			return {}

		processes = AffectedProcessesCollection()
		for child in process.children():
			processes.update(self._affecting_processes(child, packages, affected_process_factory))
			processes.update(self._affecting_children(child, packages, affected_process_factory))
		return processes



#-*- coding: utf-8 -*-
# helper.py
# Defines HelperController
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from psutil import AccessDenied

from tracer.resources.system import System
from tracer.resources.memory import dump_memory
from tracer.views.helper import HelperView
from tracer.resources.lang import _
from tracer.resources.tracer import Tracer
from tracer.resources.applications import Applications
from tracer.resources.rules import Rules


class HelperController(object):

	args = None
	packages = None

	def __init__(self, args, packages=None):
		self.args = args
		self.packages = packages

	def render(self):
		for app_name in self.args.helper:
			self.print_helper(Applications.find(app_name), self.args)
			if app_name != self.args.helper[-1]:
				print("")

	def print_helper(self, app, args):
		if app.instances:
			manager = System.package_manager()
			package = manager.provided_by(app)
			if package:
				package.load_info(System.package_manager())

			tr = Tracer(System.package_manager(), Rules, Applications)
			tr.now = self.args.now
			if self.packages:
				tr.specified_packages = self.packages

			try: affected_by = tr.trace_application(app)
			except AccessDenied: affected_by = _("You don't have enough permissions")
			affects = self._affects(app, affected_by)

			view = HelperView()
			view.assign("args", args)
			view.assign("processes", app.instances)
			view.assign("application", app)
			view.assign("package", package)
			view.assign("affected_by", affected_by)
			view.assign("affects", affects)
			view.render()
		else:
			print(_("Application called {0} is not running").format(app.name))

	def _affects(self, app, affected_by):
		if not affected_by:
			return None

		last = affected_by[-1].name()
		if last == app.name:
			return None

		if last not in [p.parent().name() for p in app.instances]:
			return None

		return last



#-*- coding: utf-8 -*-
# system.py
# Module for getting data about your operating system
# Dont worry, only necessary data required for this application.
# Tracer *will not* store, collect or send your data anywhere.
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

# WARNING: There are imports in package_manager()
import os
import pwd
import importlib
import distro
import psutil
from sys import version_info
from tracer.resources.PackageManager import PackageManager
from tracer.resources.processes import Process

class System(object):

	@staticmethod
	def distribution():
		"""
		Checks if /etc/os-release exists, and if it does, uses it to divine the name of the distribution or
		distribution like. e.g It will return 'debian' on Ubuntu systems.
		Otherwise, revert to using distro.id()
		"""
		if os.path.isfile("/etc/os-release"):
			with open("/etc/os-release") as os_release_file:
				os_release_data = {}
				distros = ["gentoo", "debian", "rhel", "centos", "ol", "mageia", "arch", "archarm", "fedora", "suse"]

				# Remove empty lines and trailing spaces
				lines = [line.rstrip() for line in os_release_file if line.rstrip()]
				for line in lines:
						os_release_key, os_release_value = line.split("=")
						os_release_data[os_release_key] = os_release_value.strip('"')
				if os_release_data["ID"] in distros:
					# Special case since centos 7 has only yum, but centos-8 uses dnf
					if os_release_data["ID"] == "centos" and os_release_data["VERSION_ID"] == 7:
						return "centos-7"
					return os_release_data["ID"]
				else:
					if "ID_LIKE" in os_release_data:
						for distribution in os_release_data["ID_LIKE"].split():
							if distribution in distros:
								return distribution
		else:
			return distro.id()

	@staticmethod
	def package_manager(**kwargs):
		"""Returns instance of package manager according to installed linux distribution"""

		def get_instance(pair):
			# WARNING: Imports here
			path, name = pair
			module = importlib.import_module(path)
			return getattr(module, name)(**kwargs)

		managers = {
			"gentoo":  [("tracer.packageManagers.portage", "Portage")],
			"debian":  [("tracer.packageManagers.dpkg", "Dpkg")],
			"rhel":  [
				("tracer.packageManagers.dnf", "Dnf5"),
				("tracer.packageManagers.dnf", "Dnf"),
				("tracer.packageManagers.yum", "Yum"),
			],
			"centos":  [
				("tracer.packageManagers.dnf", "Dnf5"),
				("tracer.packageManagers.dnf", "Dnf"),
				("tracer.packageManagers.yum", "Yum"),
			],
			"centos-7":  [("tracer.packageManagers.yum", "Yum")],
			"ol":  [
				("tracer.packageManagers.dnf", "Dnf5"),
				("tracer.packageManagers.dnf", "Dnf"),
				("tracer.packageManagers.yum", "Yum"),
			],
			"mageia":  [
				("tracer.packageManagers.dnf", "Dnf5"),
				("tracer.packageManagers.dnf", "Dnf"),
			],
			"arch":    [("tracer.packageManagers.alpm", "Alpm")],
			"archarm": [("tracer.packageManagers.alpm", "Alpm")],
			"fedora":  [
				("tracer.packageManagers.dnf", "Dnf5"),
				("tracer.packageManagers.dnf", "Dnf"),
				("tracer.packageManagers.yum", "Yum"),
			],
			"suse":  [
				("tracer.packageManagers.dnf", "Dnf5"),
				("tracer.packageManagers.dnf", "Dnf"),
			],
		}

		distribution = System.distribution()
		if distribution not in managers:
			return None

		return PackageManager(*list(map(get_instance, managers[distribution])))

	@staticmethod
	def init_system():
		"""
		Returns name of init system you are using
		e.g. init, systemd, upstart
		"""

		init = Process(1)
		name = init.name().split(" ")[0]
		return name

	@staticmethod
	def boot_time():
		# psutil-2.x.x is not backward compatible to psutil-1.x.x
		try: return psutil.boot_time()
		except AttributeError: return psutil.get_boot_time()

	@staticmethod
	def python_version():
		return "{}.{}.{}".format(version_info.major, version_info.minor, version_info.micro)

	@staticmethod
	def running_kernel_package():
		return System.package_manager().find_package(System.kernel_package_name(), os.uname()[2])

	@staticmethod
	def kernel_package_name():
		""" TODO: infer kernel package from current distribution """
		return 'kernel'

	@staticmethod
	def user():
		# getlogin is preferred because it return current username even
		# if python process is executed with sudo
		try: return os.getlogin()
		except OSError: return pwd.getpwuid(os.getuid())[0]



#-*- coding: utf-8 -*-
# applications.py
# Manager for applications file
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

from xml.dom import minidom
from xml.parsers.expat import ExpatError
from tracer.paths import DATA_DIR, USER_CONFIG_DIRS
from tracer.resources.exceptions import PathNotFound, TracerError
from tracer.resources.collections import ApplicationsCollection
from tracer.resources.lang import _
from tracer.resources.processes import Processes
from tracer.resources.system import System
from tracer.resources.SystemdDbus import SystemdDbus
from tracer.resources.pycomp import lru_cache
import os
import re


class Applications(object):

	DEFINITIONS = map(lambda x: x + "/applications.xml", [DATA_DIR] + USER_CONFIG_DIRS)

	TYPES = {
		"DAEMON"       :  "daemon",
		"STATIC"       :  "static",
		"SESSION"      :  "session",
		"APPLICATION"  :  "application",
		"ERASED"       :  "erased",
		"UNDEF"        :  "undefined" #Internal only
	}
	DEFAULT_TYPE = TYPES["APPLICATION"]
	_apps = None

	@staticmethod
	def find(app_name):
		if not Applications._apps:
			Applications._load_definitions()

		for app in Applications._apps:
			if app.name == app_name:
				return app if "rename" not in app else Applications.find(app.rename)

		# Return the default application
		return Application({
			"name": app_name,
			"type": Applications.TYPES["UNDEF"],
			"helper": None,
			"note": None,
			"ignore": False,
		})

	@staticmethod
	def all():
		if not Applications._apps:
			Applications._load_definitions()

		return Applications._apps

	@staticmethod
	def _load_definitions():
		Applications._apps = ApplicationsCollection()
		for file in Applications.DEFINITIONS:
			try: Applications._load(file)
			except PathNotFound as ex:
				if not os.path.dirname(file) in USER_CONFIG_DIRS:
					raise ex

	@classmethod
	def _load(cls, file):
		try:
			with open(file, "r") as f:
				xmldoc = minidom.parseString(f.read())
		except IOError:
			raise PathNotFound('DATA_DIR')
		except ExpatError as ex:
			msg = "Unable to parse {0}\nHint: {1}".format(file, ex)
			raise TracerError(msg)

		for applications in xmldoc.getElementsByTagName("applications"):
			cls._remove_unwanted_children(applications)
			for child in applications.childNodes:
				if child.nodeName == "app":
					attrs = dict(child.attributes.items())
					Applications._append_application(attrs)

				if child.nodeName == "group":
					cls._remove_unwanted_children(child)
					for app in child.childNodes:
						app_attrs = dict(app.attributes.items())
						group_attrs = dict(child.attributes.items())
						Applications._append_application(app_attrs, group_attrs)

	@classmethod
	def _remove_unwanted_children(cls, node):
		for child in node.childNodes:
			if child.nodeType != node.ELEMENT_NODE:
				node.removeChild(child)

	@staticmethod
	def _append_application(default_attrs, specific_attrs={}):
		application = Application(default_attrs)
		application.update(specific_attrs)

		if application in Applications._apps:
			i = Applications._apps.index(application)
			Applications._apps[i].update(application)
		else:
			application.setdefault('type', Applications.DEFAULT_TYPE)
			application.setdefault('helper', Applications._helper(application))
			application.setdefault('note', None)
			application.setdefault('ignore', False)
			Applications._apps.append(application)

	@staticmethod
	def _helper(app):
		if app.type == Applications.TYPES["DAEMON"]:
			if System.init_system() == "systemd":
				return "systemctl restart {0}".format(app.name)
			else:
				return "service {0} restart".format(app.name)

		elif app.type == Applications.TYPES["STATIC"]:
			return _("You will have to reboot your computer")

		elif app.type == Applications.TYPES["SESSION"]:
			return _("You will have to log out & log in again")

		return None


class Application:

	"""
	Represent the application defined in ``applications.xml``

	:param str name: The name of the application
	:param str type: See ``Applications.TYPES`` for possible values
	:param str helper: Describes how to restart the applications
	:param bool note: Provides additional informations to the ``helper``
	:param bool ignore: If ``True``, the application won't be printed
	:param Processes processes_factory: Class providing list of running processes

	"""

	processes_factory = Processes
	_attributes = None

	def __init__(self, attributes_dict):
		self._attributes = attributes_dict
		self._cached_name = None

	def __eq__(self, other):
		return isinstance(other, Application) and self.name == other.name

	def __ne__(self, other):
		return not self.__eq__(other)

	def __hash__(self):
		return hash(self.name)

	def __getattr__(self, item):
		return self._attributes[item]

	def __len__(self):
		return len(self._attributes)

	def __contains__(self, item):
		return item in self._attributes

	def __str__(self):
		return "<" + self.__class__.__name__ + ": " + self._attributes["name"] + ">"

	def __repr__(self):
		return self.__str__()

	def setdefault(self, key, value):
		self._attributes.setdefault(key, value)

	def update(self, values):
		if isinstance(values, Application):
			values = values._attributes
		self._attributes.update(values)

	@property
	def is_interpreted(self):
		# @TODO check all instances
		return self.instances and self.instances[0].is_interpreted

	@property
	@lru_cache(maxsize=None)
	def is_session(self):
		if self.name.startswith("ssh-") and self.name.endswith("-session"):
			return True
		return self.instances and self.instances[0].is_session

	@property
	def type(self):
		if self._attributes["type"] != Applications.TYPES["UNDEF"]:
			return self._attributes["type"]

		if self.is_session:
			return Applications.TYPES["SESSION"]

		if self.is_service:
			return Applications.TYPES["DAEMON"]

		return Applications.DEFAULT_TYPE

	@property
	@lru_cache(maxsize=None)
	def is_service(self):
		if System.init_system() == "systemd":
			return SystemdDbus().unit_path_from_id("{0}.service".format(self.name))

	# @TODO rename to helper_format
	@property
	def helper(self):
		helper = self._attributes["helper"] or Applications._helper(self)
		if System.user() != "root" and self.type == Applications.TYPES["DAEMON"]:
			if helper and not helper.startswith("sudo "):
				helper = "sudo " + helper
		return helper

	@property
	def helper_contains_formating(self):
		if not self.helper:
			return None
		return bool(re.search(r"\{.*\}", self.helper))

	@property
	def helper_contains_name(self):
		if not self.helper:
			return None
		return bool(re.search(r"\{NAME\}", self.helper))

	@property
	def helpers(self):
		"""
		Return the list of helpers which describes how to restart the application.
		When no ``helper_format`` was described, empty list will be returned.
		If ``helper_format`` contains process specific arguments such a ``{PID}``, etc.
		list will contain helper for every application instance.
		In other cases, there will be just one helper in the list.
		"""
		helpers = []
		if self.helper:
			if not self.helper_contains_formating:
				helpers.append(self.helper)
			else:
				for process in self.affected_instances:
					helpers.append(self.helper.format(
						NAME=self.name,
						PNAME=process.name,
						PID=process.pid,
						EXE=process.exe,
					))
		return helpers

	@property
	def instances(self):
		"""
		Return collection of processes with same name as application.
		I.e. running instances of the application
		"""
		return self.processes_factory.all().filtered(
			lambda process: self._attributes["name"] in [process.name(), process.real_name])

	affected_instances = None


class AffectedApplication(Application):
	@property
	def name(self):
		# We need to cache manually instead of using `@lru_cache` because this
		# property is used for self.__hash__
		if not self._cached_name:
			self._cached_name = self._name()
		return self._cached_name

	def _name(self):
		if System.init_system() == "systemd":
			bus = SystemdDbus()

			# Trying to access `self.instances` just once
			try:
				pid = self.instances[0].pid
			except IndexError:
				pid = None

			if pid:
				if self.instances and bus.unit_path_from_pid(pid):
					if not bus.has_service_property_from_pid(pid, 'PAMName'):
						Id = bus.get_unit_property_from_pid(pid, 'Id')
						if Id and re.search(r"\.service$", Id):
							return re.sub(r'\.service$', '', Id)

		if self.is_interpreted:
			return self.instances[0].real_name
		return self._attributes["name"]



from __future__ import unicode_literals

from .. import View


class RulesView(View):
	def render(self):
		line = "{0:<40}{1:<10}"
		print(line.format("Rule", "Action"))
		print(55 * "-")
		for rule in self.args.rules:
			print(line.format(rule.name, rule.action))



#-*- coding: utf-8 -*-
# lang.py
# Module working with language localizations
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#


from __future__ import absolute_import

import locale
import gettext
from gettext import NullTranslations
from sys import version_info
from tracer.paths import LANG_DIR

# python 3 compabillity settings
if version_info.major >= 3:
	# u?gettext dont exists in python3 NullTranslations
	NullTranslations.ugettext = NullTranslations.gettext
	NullTranslations.ungettext = NullTranslations.ngettext

t = gettext.translation('tracer', fallback=True, localedir=LANG_DIR)
_ = t.ugettext



#-*- coding: utf-8 -*-
# package.py
# Represents linux package
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import


class Package:
	"""Represents linux package"""

	name = None  #: Name of the package
	modified = None  #: UNIX timestamp of the modification
	description = None
	category = None
	epoch = None
	version = None
	release = None

	def __init__(self, name, modified=None):
		self.name = name
		self.modified = modified

	def __eq__(self, package):
		"""Packages are equal when they have same name"""
		return (isinstance(package, self.__class__)
			and self.name == package.name)

	def __ne__(self, package):
		return not self.__eq__(package)

	def __repr__(self):
		return "<Package:" + self.name + ">"

	def __str__(self):
		return self.name

	def __hash__(self):
		return hash(self.name)

	def load_info(self, package_manager):
		package_manager.load_package_info(self)



from __future__ import unicode_literals

from . import View
from tracer.resources.lang import _
from tracer.views.note_for_hidden import NoteForHiddenView


class InteractiveView(View):
	def render(self):
		if self.args.applications:
			print(_("You should restart:"))

		i = 1
		digits = len(str(len(self.args.applications)))
		for application in self.args.applications:
			n = "[{0}]".format(i).ljust(digits + 2)
			print("{} {}".format(n, application.name))
			i += 1

		if not self.args.args.all:
			if self.args.applications and (self.args.session_count or self.args.static_count):
				print("")

			view = NoteForHiddenView()
			view.assign("args", self.args.args)
			view.assign("total_count", self.args.total_count)
			view.assign("session_count", self.args.session_count)
			view.assign("static_count", self.args.static_count)
			view.render()




#-*- coding: utf-8 -*-
# args_parser.py
# Module for parsing console arguments
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

import argparse

try:
    import argcomplete
except ImportError:
    pass

parser = argparse.ArgumentParser(
	prog = 'tracer',
	description='Tracer finds outdated running applications in your system',
)

# Obsolete
parser.add_argument('pkgs',
	nargs='*',
	type=str,
	help='Obsolete: Use --packages instead'
)

parser.add_argument('--packages',
	dest='packages',
	nargs='*',
	type=str,
	default=[],
	help='packages that only should be traced'
)

parser.add_argument('-i', '--interactive',
	dest='interactive',
	action='store_true',
	help='run tracer in interactive mode. Print numbered applications and give helpers based on numbers'
)

parser.add_argument('-n', '--now',
	dest='now',
	action='store_true',
	help='when there are specified packages, dont look for time of their update. Use "now" instead'
)

parser.add_argument('-t', '--timestamp',
	nargs=1,
	default=[None],
	dest='timestamp',
	help='since when the updates should be'
)

parser.add_argument('-q', '--quiet',
	dest='quiet',
	action='store_true',
	help='do not print additional information'
)

parser.add_argument('-v', '--verbose',
	dest='verbose',
	action='count',
	default=0,
	help='print more informations. Use -v or -vv'
)

parser.add_argument('-s', '--show',
	nargs='+',
	dest='helper',
	metavar='app_name',
	help='show helper for given application'
)

parser.add_argument('--helpers',
	dest='helpers',
	action='store_true',
	help='not list applications, but list their helpers'
)

parser.add_argument('-a', '--all',
	dest='all',
	action='store_true',
	help='list even session and unrestartable applications'
)

parser.add_argument('--hooks-only',
	dest='hooks_only',
	action='store_true',
	help='do not print traced applications, only run their hooks'
)

filter_type = parser.add_mutually_exclusive_group()
filter_type.add_argument('--reboot-only',
	dest='reboot_only',
	action='store_true',
	help='list only items requiring a reboot'
)

filter_type.add_argument('--daemons-only', '--services-only',
	dest='daemons_only',
	action='store_true',
	help='list only daemons/services'
)

parser.add_argument('--version',
	dest='version',
	action='store_true',
	help='print program version'
)

parser.add_argument('--show-resource',
	nargs=1,
	choices=['packages', 'processes', 'rules', 'applications', 'system'],
	dest='resource',
	help='provide informations about selected resource'
)

user = parser.add_mutually_exclusive_group()
user.add_argument("-u", "--user",
	nargs=1,
	dest='user',
	metavar='username'
)

user.add_argument("-r", "--root",
	dest='user',
	action="store_const",
	const='root'
)

user.add_argument("-e", "--everyone",
	dest='user',
	action="store_const",
	const='*'
)

user.add_argument("--erased",
	dest="erased",
	action="store_true",
	help="print even section with erased packages (DNF only)",
	default=False
)
try:
    argcomplete.autocomplete(parser)
except NameError:
    pass



#-*- coding: utf-8 -*-
# router.py
# Router chooses the right controller and its method and calls it
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

import os
from tracer.version import __version__
from tracer.resources.lang import _
# WARNING: There are local imports in dispatch method


class Router:

	args = None
	packages = None

	def __init__(self, args, packages):
		self.args = args
		self.packages = packages

	def dispatch(self):
		if self.args.helper:
			from tracer.controllers.helper import HelperController
			controller = HelperController(self.args, self.packages)
			controller.render()

		elif self.args.version:
			print(__version__)

		elif self.args.resource:
			from tracer.controllers.resource import ResourceController
			controller = ResourceController(self.args)
			controller.render()

		else:
			from tracer.controllers.default import DefaultController
			controller = DefaultController(self.args, self.packages)
			if self.args.helpers:
				controller.render_helpers()
			elif self.args.interactive:
				controller.render_interactive()
			else:
				controller.render()



from __future__ import absolute_import

from tracer.query import Query
from tracer.resources.package import Package
from tracer.resources.applications import Application
from tracer.resources.processes import Process

__all__ = [
	Query,
	Package,
	Application,
	Process,
]


#-*- coding: utf-8 -*-
# alpm.py
#
# Copyright (C) 2016 Patrick Griffis <tingping@tingping.se>
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

from tracer.resources.system import System
if System.distribution() in ["arch", "archarm"]:

	import bisect
	from .ipackageManager import IPackageManager
	from tracer.resources.package import Package
	from tracer.resources.collections import PackagesCollection
	from tracer.resources.applications import Applications
	import pyalpm

	class Alpm(IPackageManager):

		def __init__(self, *args, **kwargs):
			self.opts = kwargs
			self.handle = pyalpm.Handle('/', '/var/lib/pacman')
			self.db = self.handle.get_localdb()

		def packages_newer_than(self, unix_time):
			"""
			Returns list of packages which were modified between unix_time and present
			"""
			new_pkgs = PackagesCollection()
			for pkg in self.db.pkgcache:
				if pkg.installdate > unix_time:
					new_pkgs.append(Package(pkg.name, pkg.installdate))

			return new_pkgs

		def package_files(self, pkg_name):
			"""
			Returns list of files provided by package
			"""
			pkg = self.db.get_pkg(pkg_name)
			if not pkg:
				return []

			return ['/' + f[0] for f in pkg.files]

		def load_package_info(self, package):
			"""
			From database load informations about given package and set them to it
			"""
			if not package:
				return

			pkg = self.db.get_pkg(package.name)
			if not pkg:
				return

			package.description = pkg.desc
			package.modified = pkg.installdate
			# Don't have categories

		def provided_by(self, app):
			"""
			Returns name of package which provides given application
			"""
			# We need a full path to the binary
			process = app.instances[0]
			return self._file_provided_by(process.exe)

		def find_package(self, pkg_name, version):
			"""
			Find a package by name and some other input criteria
			"""
			pkg = self.db.get_pkg(pkg_name)
			if pkg and pyalpm.vercmp(pkg.version, version) == 0:
				return pkg

		def compare_packages(self, package1, package2):
			"""
			vercmp returns:
			< 0 if ver1 < ver2
			0 if ver1 == ver2
			> 0 if ver1 > ver2
			"""
			return pyalpm.vercmp(package1.version, package2.version)

		@staticmethod
		def _bsearch_list(l, item):
			"""
			Searches a sorted list, returns True if found
			"""
			i = bisect.bisect_left(l, item)
			if i != len(l) and l[i] == item:
				return True
			else:
				return False

		def _file_provided_by(self, file_name):
			"""
			Returns name of package which provides given file
			"""
			# This is a bit slow
			for pkg in self.db.pkgcache:
				files = [f[0] for f in pkg.files]
				if self._bsearch_list(files, file_name[1:]):
					return Package(pkg.name, pkg.installdate)
			else:
				return None




#-*- coding: utf-8 -*-
# ipackageManager.py
# This class should be inherited by any other package manager
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#


class IPackageManager(object):

	def __init__(self, **kwargs):
		"""This class is 'interface' so you can't create an instance"""
		raise NotImplementedError

	def packages_newer_than(self, unix_time):
		"""
		Returns list of packages which were modified between unix_time and present
		Packages in list should be dictionaries with keys {"name", "modified"}
		"""
		raise NotImplementedError

	def package_files(self, pkg_name):
		"""Returns list of files provided by package"""
		raise NotImplementedError

	def load_package_info(self, package):
		"""From database load informations about given package and set them to it"""
		raise NotImplementedError

	def provided_by(self, app):
		"""Returns name of package which provides given application"""
		raise NotImplementedError

	def find_package(self, pkg_name, search):
		"""Find a package by name and some other input criteria"""
		raise NotImplementedError

	def compare_packages(self, package1, package2):
		"""
		Compares two packages by their version information
		Returns:
		0 if they are equal
		1 if package1 > package2
		-1 if package2 > package1
		"""
		raise NotImplementedError

	@staticmethod
	def _pkg_name_without_version(pkg_name):
		try:
			pkg_name = pkg_name[:pkg_name.index('.')]  # Cut from first . to end
			pkg_name = pkg_name[:pkg_name.rindex('-')] # Cut from last  - to end
		except ValueError: pass
		return pkg_name



#-*- coding: utf-8 -*-
# default.py
# Defines DefaultController
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

import os
import sys
from tracer.hooks import HooksObserver
from tracer.views.default import DefaultView
from tracer.views.interactive import InteractiveView
from tracer.views.note_for_hidden import NoteForHiddenView
from tracer.resources.lang import _
from tracer.resources.tracer import Tracer
from tracer.resources.system import System
from tracer.resources.memory import dump_memory
from tracer.resources.applications import Applications
from tracer.resources.rules import Rules
from tracer.controllers.helper import HelperController


# compatibility with Py2 and Py3 - rename raw_input() to input() on Py2
try:
	input = raw_input
except NameError:
	pass


class DefaultController(object):

	args = None
	tracer = None
	processes = None

	def __init__(self, args, packages):
		self.args = args
		self.tracer = Tracer(
			System.package_manager(erased=args.erased),
			Rules,
			Applications,
			memory=dump_memory,
			hooks_observer=HooksObserver(),
			erased=args.erased
		)
		self.tracer.now = args.now
		self.tracer.timestamp = args.timestamp[0]
		if packages:
			self.tracer.specified_packages = packages

		self.applications = self.tracer.trace_affected(self._user(args.user))
		if self.args.daemons_only:
			self.applications = self.applications.filter_types([Applications.TYPES["DAEMON"]])
		elif self.args.reboot_only:
			self.applications = self.applications.filter_types([Applications.TYPES["STATIC"]])

	def render(self):
		if not self.args.hooks_only:
			view = DefaultView()
			view.assign("applications", self.applications)
			view.assign("args", self.args)
			view.render()
			self.create_reboot_required_file()
		exit(self.status_code())

	def render_helpers(self):
		helper_controller = HelperController(self.args)
		for application in self._restartable_applications(self.applications, self.args):
			helper_controller.print_helper(application, self.args)
			print("")

		view = NoteForHiddenView()
		view.assign("args", self.args)
		view.assign("total_count", len(self.applications))
		view.assign("session_count", self.applications.count_type(Applications.TYPES['SESSION']))
		view.assign("static_count", self.applications.count_type(Applications.TYPES['STATIC']))
		view.render()

	def render_interactive(self):
		helper_controller = HelperController(self.args)
		filtered = self._restartable_applications(self.applications, self.args).sorted("name")

		while True:
			view = InteractiveView()
			view.assign("applications", filtered)
			view.assign("args", self.args)
			view.assign("total_count", len(self.applications))
			view.assign("session_count", self.applications.count_type(Applications.TYPES['SESSION']))
			view.assign("static_count", self.applications.count_type(Applications.TYPES['STATIC']))
			view.render()

			# If there are only hidden applications (any listed)
			if view.get("total_count") == view.get("session_count") + view.get("static_count"):
				break

			print("\n" + _("Press application number for help or 'q' to quit"))
			answer = input("--> ")
			try:
				if answer == "q": return
				elif int(answer) <= 0 or int(answer) > len(filtered): raise IndexError
				helper_controller.print_helper(filtered[int(answer) - 1], self.args)

			except (SyntaxError, IndexError, ValueError):
				print(_("Wrong application number"))

			sys.stdout.write("\n-- " + _("Press <enter> to get list of applications") + " --")
			input()

	def status_code(self):
		"""
		0   - No affected applications
		101 - Found some affected applications
		102 - Found some affected daemons
		103 - Session restart needed
		104 - Reboot needed
		"""
		code = 0
		if len(self.applications) > 0:
			code = 101

		if self.applications.count_type(Applications.TYPES['DAEMON']):
			code = 102

		if self.applications.count_type(Applications.TYPES['SESSION']):
			code = 103

		if self.applications.count_type(Applications.TYPES['STATIC']):
			code = 104
		return code

	def create_reboot_required_file(self):
		"""
		If a reboot is needed, create a /run/reboot-required file.
		This is how Debian/Ubuntu distros does it.
		"""
		if os.getuid() == 0 and self.applications.count_type(Applications.TYPES["STATIC"]):
			with open("/run/reboot-required", "w") as fp:
				fp.write("Tracer says reboot is required for:\n")
				for app in self.applications.filter_types([Applications.TYPES["STATIC"]]).unique().sorted("name"):
					fp.write(f"- {app.name}\n")

	def _restartable_applications(self, applications, args):
		return applications.exclude_types([
			Applications.TYPES['STATIC'],
			Applications.TYPES['SESSION']
		]) if not args.all else applications

	def _user(self, user):
		if   user == ['*'] or user == '*':    return None
		elif user == 'root': return user
		elif not user:       return System.user()
		else: return user[0]



#-*- coding: utf-8 -*-
# main.py
# Tracer finds outdated running applications in your system
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import print_function

import os
import sys
import time
from tracer.resources.router import Router
from tracer.resources.args_parser import parser
from tracer.resources.package import Package
from tracer.resources.exceptions import (
	TracerError,
	UnsupportedDistribution,
	PathNotFound,
	LockedDatabase,
	DatabasePermissions,
)
from tracer.resources.lang import _


def run():
	args = parser.parse_args()

	# If there is something on stdin (that means piped into tracer)
	stdin_packages = []
	if not sys.stdin.isatty() and "SSH_CONNECTION" not in os.environ:
		stdin_packages = sys.stdin.readline().split()

	# All input packages enchanced by actual time (as modified time)
	packages = []
	for package in args.packages + args.pkgs + stdin_packages:
		packages.append(Package(package, time.time() if args.now else None))

	try:
		router = Router(args, packages)
		return router.dispatch()

	except (UnsupportedDistribution, PathNotFound, LockedDatabase) as ex:
		ex.print()
		exit(1)
	except DatabasePermissions as ex:
		ex.print()
		print(_("You will probably need to run tracer as root"))
		exit(1)
	except TracerError as ex:
		ex.print()
		exit(1)
	except (KeyboardInterrupt, EOFError):
		print("")



#-*- coding: utf-8 -*-
# dnf.py
# Module to work with DNF package manager class
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

import os.path

from tracer.resources.system import System
if System.distribution() in ["rhel", "fedora", "centos", "centos-7", "mageia", "suse", "ol"]:

	import subprocess
	from tracer.packageManagers.rpm import Rpm
	from tracer.resources.package import Package
	from tracer.resources.collections import PackagesCollection

	class Dnf(Rpm):

		def __init__(self, **kwargs):
			super(Dnf, self).__init__(**kwargs)
			if os.path.exists('/usr/lib/sysimage/dnf/history.sqlite'):
				self.opts['sysimage_persistdir'] = True
			if os.path.exists('/usr/lib/sysimage/dnf/history.sqlite') or os.path.exists('/var/lib/dnf/history.sqlite'):
				self.opts['modern_swdb'] = True

		@property
		def history_path(self):
			if self.opts.get('modern_swdb'):
				if self.opts.get('sysimage_persistdir'):
					return '/usr/lib/sysimage/dnf/history.sqlite'
				else:
					return '/var/lib/dnf/history.sqlite'
			if self.opts.get('sysimage_persistdir'):
				return '/usr/lib/sysimage/dnf/history/'
			else:
				return '/var/lib/dnf/history/'

		def package_files(self, pkg_name):
			if not self.opts.get("erased"):
				return super(Dnf, self).package_files(pkg_name)

			# TODO Running an external command is tooooo slow, use python API instead
			process = subprocess.Popen(["dnf", "repoquery", "-q", "-l", pkg_name], stdout=subprocess.PIPE)
			out = process.communicate()[0]
			return out.decode().split("\n")


	class FakeDnf5(Dnf):
		def packages_newer_than(self, unix_time):
			return []

		def package_files(self, pkg_name):
			return []


	class Dnf5(Dnf):
		def __new__(cls, *args, **kwargs):
			"""
			We are going to enable DNF5 on all Fedora, EPEL, Mageia, openSUSE,
			etc systems and not all of them may have DNF5 (yet).
			"""
			try:
				import libdnf5
				return super().__new__(cls)
			except ImportError:
				return FakeDnf5(*args, **kwargs)

		def packages_newer_than(self, unix_time):
			import libdnf5
			base = libdnf5.base.Base()
			base.setup()

			# Package names as a keys and timestamps as values
			packages = {}

			history = base.get_transaction_history()
			for transaction in history.list_all_transactions():
				if transaction.get_dt_start() < unix_time:
					continue
				for package in transaction.get_packages():
					packages[package.get_name()] = transaction.get_dt_start()

			collection = PackagesCollection()
			for name, timestamp in packages.items():
				collection.append(Package(name, timestamp))

			return collection

		def package_files(self, pkg_name):
			if not self.opts.get("erased"):
				return super(Dnf, self).package_files(pkg_name)

			import libdnf5
			base = libdnf5.base.Base()
			base.setup()

			repo_sack = base.get_repo_sack()
			if hasattr(repo_sack, "load_repos"):
				# On F41+, this is what should be done
				repo_sack.load_repos()
			else:
				# On F40 there is only python3-libdnf5-5.1.17 so we have to use
				# this now-deprecated function
				repo_sack.update_and_load_enabled_repos(True)

			query = libdnf5.rpm.PackageQuery(base)
			query.filter_name([pkg_name])
			paths = []
			for pkg in query:
				paths.extend(pkg.get_files())
			return paths



from __future__ import unicode_literals

from .. import View


class ApplicationsView(View):
	def render(self):
		line = "{0:<40}{1:<20}{2:<10}{3:<50}"
		print(line.format("Application", "Type", "Ignore", "Helper"))
		print(120 * "-")
		for application in self.args.applications:
			print(line.format(application.name, application.type, application.ignore, application.helper or ""))



# -*- coding: utf-8 -*-
# hooks.py
# Module providing hooks functionality
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

import os
from tracer.paths import HOOKS_DIRS
from tracer.resources.pycomp import load_source

_hooks = {}


class HooksObserver(object):
	"""
	Provides interface for calling user hooks

	When initializing, loads all hooks located in HOOKS_DIRS. Then it can called
	with application name as argument. Observer ensures calling all hooks defined for the application.
	"""
	def __init__(self):
		_register_hooks()

	def __call__(self, app):
		if app not in _hooks:
			return

		[f() for f in _hooks[app]]


def match(apps):
	"""
	Decorator for tracer hooks.

	Example::

		from tracer import hooks

		@hooks.match("foo")
		def hook_app():
		    print("Hey, application foo was found")

	.. note::
		You can match multiple applications by calling ``@hooks.match`` with list of them.

	"""
	def decorator(f):
		for app in apps if type(apps) == list else [apps]:
			if app not in _hooks:
				_hooks[app] = []
			_hooks[app].append(f)
		return f
	return decorator


def _register_hooks():
	# see search function here
	# http://stackoverflow.com/a/4788700/3285282

	for hook_dir in HOOKS_DIRS:
		for root, dirs, files in os.walk(hook_dir):
			for fname in files:
				modname = os.path.splitext(fname)[0]
				try:
					load_source(modname, os.path.join(root, fname))
				except Exception:
					continue



#-*- coding: utf-8 -*-
# collections.py
# Define various kind of collections
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

from functools import cmp_to_key
from psutil import NoSuchProcess


class Collection(list):

	def replace_values(self, attribute, source_value, required_value):
		for app in self:
			if getattr(app, attribute) == source_value:
				app.update({attribute: required_value})


	def sorted(self, attribute):
		def _sort(app1, app2):
			value1 = _value(app1, attribute)
			value2 = _value(app2, attribute)

			# Make sure the None values are at the end
			# This shouldn't happen anyway but sometimes if does,
			# for some reason, see #151 or #156
			if not value1:
				return 1

			if not value2:
				return -1

			# https://stackoverflow.com/a/13239857/3285282
			if value1 < value2:
				return -1
			if value1 > value2:
				return 1
			return 0

		def _value(app, attribute):
			if callable(getattr(self[0], attribute)):
				return getattr(app, attribute)()
			return getattr(app, attribute)

		return sorted(self, key=cmp_to_key(_sort))


class ApplicationsCollection(Collection):

	def with_helpers(self):
		applications = filter(lambda app: app.helper, self)
		return ApplicationsCollection(applications)

	def without_helpers(self):
		applications = filter(lambda app: not app.helper, self)
		return ApplicationsCollection(applications)

	def exclude_types(self, app_types):
		"""app_types -- see Applications.TYPES"""
		applications = filter(lambda app: app.type not in app_types, self)
		return ApplicationsCollection(applications)

	def filter_types(self, app_types):
		"""app_types -- see Applications.TYPES"""
		applications = filter(lambda app: app.type in app_types, self)
		return ApplicationsCollection(applications)

	def count_type(self, app_type):
		count = 0
		for application in self:
			if application.type == app_type:
				count += 1
		return count

	def unique(self):
		return ApplicationsCollection(set(self))


class ProcessesCollection(Collection):

	def owned_by(self, user):
		if not user:
			return self
		return self.filtered(lambda process: process.username() == user)

	def newer_than(self, timestamp):
		return self.filtered(lambda process: process.create_time() >= timestamp)

	def unique(self):
		unique = set()
		for process in self:
			try: unique.add(process)
			except NoSuchProcess: pass
		return ProcessesCollection(unique)

	def filtered(self, function):
		processes = ProcessesCollection()
		for process in self:
			try:
				if function(process):
					processes.append(process)
			except NoSuchProcess: pass
		return processes


class AffectedProcessesCollection(ProcessesCollection):

	def update(self, iterable):
		for x in iterable:
			if x in self:
				self[self.index(x)].update(x)
			else:
				self.append(x)


class PackagesCollection(Collection):

	_package_manager = None

	def __init__(self, *args):
		list.__init__(self, *args)

	def intersection(self, packages):
		if packages is not None:
			return PackagesCollection(set(packages).intersection(self))
		return self

	@property
	def files(self):
		files = []
		for package in self:
			files.extend(self._package_manager.package_files(package.name))
		return set(files)

	def unique_newest(self):
		packages = {}
		for p in self:
			if p.name in packages:
				if packages[p.name].modified > p.modified:
					continue
			packages[p.name] = p
		return PackagesCollection(packages.values())



#-*- coding: utf-8 -*-
# resource.py
# Defines ResourceController
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#


from __future__ import absolute_import

import os
import psutil
from datetime import datetime
from tracer.resources.processes import Processes
from tracer.resources.system import System
from tracer.resources.rules import Rules
from tracer.resources.applications import Applications
from tracer.version import __version__

from tracer.views.resource.processes import ProcessesView
from tracer.views.resource.packages import PackagesView
from tracer.views.resource.rules import RulesView
from tracer.views.resource.applications import ApplicationsView
from tracer.views.resource.system import SystemView


class ResourceController(object):

	args = None

	def __init__(self, args):
		self.args = args

	def render(self):
		r = self.args.resource[0]
		if   r == 'processes':    self.render_processes()
		elif r == 'packages':     self.render_packages()
		elif r == 'rules':        self.render_rules()
		elif r == 'applications': self.render_applications()
		elif r == 'system':       self.render_system()

	def render_processes(self):
		view = ProcessesView()
		view.assign('processes', Processes.all())
		view.render()

	def render_packages(self):
		timestamp = self.args.timestamp[0] if self.args.timestamp[0] else System.boot_time()
		manager = System.package_manager()
		packages = manager.packages_newer_than(timestamp)

		view = PackagesView()
		view.assign('packages', packages)
		view.assign('boot_time', System.boot_time())
		view.render()

	def render_rules(self):
		view = RulesView()
		view.assign('rules', Rules.all())
		view.render()

	def render_applications(self):
		view = ApplicationsView()
		view.assign('applications', Applications.all())
		view.render()

	def render_system(self):
		uptime = datetime.now() - datetime.fromtimestamp(System.boot_time())
		uptime = str(uptime).split('.')[0]

		try:
			users = set([user.name for user in psutil.get_users()])
		except AttributeError:
			users = set([user.name for user in psutil.users()])
		package_managers = System.package_manager().names()

		view = SystemView()
		view.assign('python', System.python_version())
		view.assign('distribution', System.distribution())
		view.assign('package_managers', package_managers)
		view.assign('init', System.init_system())
		view.assign('uptime', uptime)
		view.assign('user', System.user())
		view.assign('users', users)
		view.assign('version', __version__)
		view.assign('rules_count', len(Rules.all()))
		view.assign('applications_count', len(Applications.all()))
		view.render()



#-*- coding: utf-8 -*-
# dpkg.py
# Module to work with dpkg based package managers
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import


from tracer.resources.system import System
if System.distribution() == "debian":

	from .ipackageManager import IPackageManager
	from tracer.resources.package import Package
	from tracer.resources.collections import PackagesCollection
	import subprocess
	import time
	import os

	class Dpkg(IPackageManager):

		"""
		Package manager class - DPKG
		"""

		# noinspection PyMissingConstructor
		def __init__(self, **kwargs):
			self.opts = kwargs

		@property
		def dpkg_log(self): return '/var/log/dpkg.log'

		def packages_newer_than(self, unix_time):
			"""
			Returns list of packages which were modified between unix_time and present
			Requires root permissions.
			"""
			newer = PackagesCollection()
			log = open(self.dpkg_log, 'r')
			for line in log:
				line = line.split(" ")

				if line[2] != "upgrade":
					continue

				# There actually should be %e instead of %d
				modified = time.mktime(
					time.strptime(line[0] + " " + line[1],
					"%Y-%m-%d %H:%M:%S"))

				if modified >= unix_time:
					pkg_name = line[3].split(":")[0]
					newer.append(Package(pkg_name, modified))
			return newer

		def package_files(self, pkg_name):
			"""Returns list of files provided by package"""
			files = []
			fnull = open(os.devnull, 'w')
			command = ['dpkg-query', '-L', pkg_name]
			process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=fnull)
			out = process.communicate()[0]
			for file in out.decode().split('\n')[:-1]:
				if os.path.isfile(file):
					files.append(file)
			return files

		def load_package_info(self, package):
			"""From database load informations about given package and set them to it"""
			description = None

			process = subprocess.Popen(['dpkg', '-s', package.name], stdout=subprocess.PIPE)
			out = process.communicate()[0]
			out = out.decode().split('\n')

			for line in out:
				if line.startswith("Description:"):
					description = line.split("Description:")[1].strip()

			package.description = description

		def provided_by(self, app):
			"""Returns name of package which provides given application"""
			command = ['dlocate', '-S', app.name]
			process = subprocess.Popen(command, stdout=subprocess.PIPE)
			pkg_name = process.communicate()[0]
			pkg_name = pkg_name.decode().split('\n')[0]
			return Package(pkg_name.split(':')[0])



from __future__ import print_function
from __future__ import unicode_literals

from . import View
from tracer.resources.lang import _


class HelperView(View):
	def render(self):

		self.print("* {app_name}".format(app_name=self.args.application.name))

		# Package informations
		if self.args.package:
			self.print("    Package:     {pkg_name}"        .format(pkg_name=self.args.package.name))
			self.print("    Description: {pkg_description}" .format(pkg_description=self.args.package.description))
			self.print("    Type:        {type}"            .format(type=self.args.application.type.capitalize()))
			if self.args.application.affected_instances:
				self.print("    Executable:  {executable}".format(executable=self.args.application.affected_instances[0].exe))
		else:
			self.print("    Package:     {app_name} is not provided by any package"
				.format(app_name=self.args.application.name))

		# State
		indent = "    State:       "
		i = 0
		for process in self.args.processes:
			self.print(indent + "{app_name} has been started by {user} {time} ago. PID - {pid}".format(
					app_name=self.args.application.name,
					user=process.username(),
					time=process.str_started_ago,
					pid=process.pid
			))
			indent = "                 "
			i += 1
			if i >= 3:
				self.print("                 ...")
				break

		# Affected by
		if self.args.args.verbose > 0:
			self.print("")
			self.render_affected_by()

		# How to restart
		if self.args.application.helper or self.args.affects:
			self.print("")
			self.print("    {title}:".format(title=_('How to restart')))
			if not self.args.affected_by:
				self.print("        {app_name} does not need restarting".format(app_name=self.args.application.name))
			elif self.args.affects:
				self.print("        " + _("It's a part of application called {0}").format(self.args.affects))
			else:
				for helper in self.args.application.helpers:
					self.print("        {how_to_restart}".format(how_to_restart=helper))

				if self.args.application.note:
					self.print("\n       - " + self.args.application.note)

	def render_affected_by(self):
		default_level = 2
		indent = "    "
		self.print(indent + _("Affected by") + ":")

		if type(self.args.affected_by) == str:
			self.print(default_level * indent + self.args.affected_by)
			return

		printed_packages = set()
		for process in self.args.affected_by:
			indent_level = default_level

			if process not in self.args.processes:
				self.print(indent_level * indent + "{0} ({1})".format(process.name(), process.pid))
				indent_level += 1

			for package in process.packages:
				if package.name not in printed_packages or indent_level > 2:
					self.print(indent_level * indent + package.name)
					printed_packages.add(package.name)

				if self.args.args.verbose < 2:
					continue

				indent_level += 1
				for file in package.files:
					self.print(indent_level * indent + file)

				indent_level -= 1



#-*- coding: utf-8 -*-
# rpm.py
# Base RPM package manager class
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import


from tracer.resources.system import System
if System.distribution() in ["fedora", "rhel", "centos", "centos-7", "mageia", "ol"]:

	from os import listdir
	from .ipackageManager import IPackageManager
	from tracer.resources.package import Package
	from tracer.resources.collections import PackagesCollection
	from tracer.resources.exceptions import LockedDatabase, DatabasePermissions
	from tracer.resources.pycomp import PY3
	import sqlite3
	import rpm
	import os

	class Rpm(IPackageManager):

		"""
		Package manager class - RPM
		"""

		# noinspection PyMissingConstructor
		def __init__(self, **kwargs):
			self.opts = kwargs

		@property
		def history_path(self): return NotImplemented

		def packages_newer_than(self, unix_time):
			"""
			Returns list of packages which were modified between unix_time and present
			Requires root permissions.
			"""

			# Package manager wasn't used yet
			if not os.path.exists(self.history_path):
				return PackagesCollection([])

			if self.opts.get('modern_swdb'):
				sql = """
					SELECT DISTINCT rpm.name, trans.dt_end AS end

					FROM trans JOIN trans_item JOIN rpm
					ON trans.id=trans_item.trans_id
					AND trans_item.item_id=rpm.item_id

					WHERE trans.dt_begin > ?
					ORDER BY rpm.name
				"""
			else:
				sql = """
					SELECT DISTINCT pkgtups.name, trans_end.timestamp AS end

					FROM trans_beg JOIN trans_end JOIN trans_data_pkgs JOIN pkgtups
					ON trans_beg.tid=trans_end.tid
					AND trans_data_pkgs.tid=trans_beg.tid
					AND trans_data_pkgs.pkgtupid=pkgtups.pkgtupid

					WHERE trans_beg.timestamp > ?
					ORDER BY pkgtups.name
				"""

			try:
				packages = PackagesCollection()
				sqlite = self._database_file()
				conn = sqlite3.connect('file:' + sqlite + '?mode=ro', uri=True)
				conn.row_factory = sqlite3.Row
				cursor = conn.cursor()
				cursor.execute(sql, [unix_time])

				for result in cursor.fetchall():
					packages.append(Package(result['name'], result['end']))

				return packages

			except sqlite3.OperationalError as e:
				raise LockedDatabase() if str(e) == 'database is locked' else DatabasePermissions()

		def package_files(self, pkg_name):
			"""
			Returns list of files provided by package
			See also: http://docs.fedoraproject.org/en-US/Fedora_Draft_Documentation/0.1/html/RPM_Guide/ch04s02s03.html
			"""
			ts = rpm.TransactionSet()
			mi = ts.dbMatch("name", pkg_name)
			packages = list(mi)

			# Tracer will not find uninstalled applications
			if not packages:
				return []

			if PY3:
				files = rpm.files(packages[0])
				return [x.name for x in files]
			else:
				files = rpm.fi(packages[0])
				return [f[0] for f in files]

		def find_package(self, name, evra):
			evra = self._splitEvra(evra)
			ts = rpm.TransactionSet()
			mi = ts.dbMatch("name", name)

			for hdr in mi:
				if hdr[rpm.RPMTAG_EPOCH] == evra[0] and hdr[rpm.RPMTAG_VERSION] == evra[1] and hdr[rpm.RPMTAG_RELEASE] == evra[2] and hdr[rpm.RPMTAG_ARCH] == evra[3]:
					package = Package(name)
					self._load_package_info_from_hdr(package, hdr)

					return package

			return None

		def load_package_info(self, package):
			"""From database load informations about given package and set them to it"""
			if not package:
				return None

			ts = rpm.TransactionSet()
			mi = ts.dbMatch("name", package.name)

			""" Find the latest one if there are multiple versions"""
			latest = None
			for hdr in mi:
				if latest is None:
					latest = hdr
				else:
					compare = rpm.labelCompare((str(latest[rpm.RPMTAG_EPOCH]), str(latest[rpm.RPMTAG_VERSION]), str(latest[rpm.RPMTAG_RELEASE])),
						(str(hdr[rpm.RPMTAG_EPOCH]), str(hdr[rpm.RPMTAG_VERSION]), str(hdr[rpm.RPMTAG_RELEASE])))

					if compare == -1:
						latest = hdr

			if latest is None:
				return

			self._load_package_info_from_hdr(package, latest)

		def compare_packages(self, p1, p2):
			"""
			labelCompare returns:
			0 if the EVR matches
			1 if EVR(1) > EVR(2)
			-1 if EVR(2) > EVR(1)
			"""
			return rpm.labelCompare((str(p1.epoch), str(p1.version), str(p1.release)), (str(p2.epoch), str(p2.version), str(p2.release)))

		def provided_by(self, app):
			"""Returns name of package which provides given application"""
			# `rpm -qf ...` needs full path to binary, not only its name
			process = app.instances[0]  # @TODO Reimplement for all processes
			package = self._file_provided_by(process.exe)
			if package:
				# If package is interpreter, return the package providing that interpreted file
				if package.category == 'Development/Languages':
					for arg in process.cmdline()[1:]:
						if os.path.isfile(arg):
							package = self._file_provided_by(arg)
							return package if package else None
				return package
			return None

		def _splitEvra(self, evra):
			"""
			Derived from rpmUtils.miscutils.splitFilename
			https://github.com/rpm-software-management/yum/blob/master/rpmUtils/miscutils.py

			Given: 9-123a.ia64
			Return: (9, 123a, 1, ia64)
			"""

			archIndex = evra.rfind('.')
			arch = evra[archIndex + 1:]

			relIndex = evra[:archIndex].rfind('-')
			rel = evra[relIndex + 1:archIndex]

			verIndex = evra[:relIndex].rfind('-')
			ver = evra[verIndex + 1:relIndex]

			epochIndex = evra.find(':')
			if epochIndex == -1:
				epoch = None
			else:
				epoch = evra[:epochIndex]

			return epoch, ver, rel, arch

		def _load_package_info_from_hdr(self, package, hdr):
			package.description = hdr[rpm.RPMTAG_SUMMARY]
			package.category = hdr[rpm.RPMTAG_GROUP]

			epoch = hdr[rpm.RPMTAG_EPOCH]
			if epoch:
				package.epoch = epoch

			package.version = hdr[rpm.RPMTAG_VERSION]
			package.release = hdr[rpm.RPMTAG_RELEASE]

		def _file_provided_by(self, file):
			"""Returns name of package which provides given file"""
			ts = rpm.TransactionSet()
			db = ts.dbMatch("basenames", file)
			if db.count() == 0:
				return None

			pkg = next(db)
			p = Package(pkg[rpm.RPMTAG_NAME])
			p.category = pkg[rpm.RPMTAG_GROUP]
			return p

		def _database_file(self):
			"""Returns path to yum history database file"""
			if self.opts.get('modern_swdb'):
				return self.history_path
			for file in sorted(listdir(self.history_path), reverse=True):
				if file.startswith("history-") and file.endswith(".sqlite"):
					return self.history_path + file



from __future__ import unicode_literals

from .. import View


class ProcessesView(View):
	def render(self):
		line = "{0:<10}{1:<20}{2:<20}{3:<10}"
		print(line.format("PID", "Time running", "User", "Process"))
		print(100 * "-")
		for process in self.args.processes.sorted("create_time"):
			print(line.format(process.pid, process.str_started_ago, process.username(), process.name()))



from __future__ import unicode_literals

from .. import View


class SystemView(View):
	def render(self):
		line = "{0:<20}{1}"
		print(line.format("Python:", self.args.python))
		print(line.format("Distribution:", self.args.distribution))
		print(line.format("Package Manager:", ", ".join(self.args.package_managers)))
		print(line.format("Init system:", self.args.init))
		print(line.format("Uptime:", self.args.uptime))
		print(line.format("User:", self.args.user))

		if len(self.args.users) > 1:
			print(line.format("Users:", self.args.users))

		print("")
		print(line.format("Tracer version:", self.args.version))
		print(line.format("Rules:", self.args.rules_count))
		print(line.format("Applications:", self.args.applications_count))



#-*- coding: utf-8 -*-
# portage.py
# Module to work with portage package manager class
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import
from __future__ import unicode_literals


from tracer.resources.system import System
if System.distribution() == "gentoo":

	from .ipackageManager import IPackageManager
	from tracer.resources.package import Package
	from tracer.resources.collections import PackagesCollection
	from gentoolkit.helpers import FileOwner
	from tracer.resources.exceptions import DatabasePermissions
	from tracer.resources.applications import Applications
	import os
	import portage
	import subprocess
	import time

	class Portage(IPackageManager):

		"""
		Package manager class - Portage
		"""

		# noinspection PyMissingConstructor
		def __init__(self, **kwargs):
			self.opts = kwargs

		def packages_newer_than(self, unix_time):
			"""
			Returns list of packages which were modified between unix_time and present
			Requires root permissions.
			"""

			if not os.access('/var/log/emerge.log', os.R_OK):
				raise DatabasePermissions

			newer = PackagesCollection()
			process = subprocess.Popen(['qlop', '-lC'], stdout=subprocess.PIPE)
			packages = process.communicate()[0]
			for package in packages.decode().split('\n')[:-1]:
				package = package.split(" >>> ")

				# There actually should be %e instead of %d
				modified = time.mktime(time.strptime(package[0], "%a %b %d %H:%M:%S %Y"))
				if modified >= unix_time:
					pkg_name = package[1] # Package name with version, let's cut it off
					pkg_name = self._pkg_name_without_version(pkg_name)
					newer.append(Package(pkg_name, modified))

			return newer

		def package_files(self, pkg_name):
			"""Returns list of files provided by package"""
			vartree = portage.db[portage.root]['vartree']
			cpv = str(vartree.dep_bestmatch(pkg_name))

			contents = vartree.dbapi.aux_get(cpv, ['CONTENTS'])[0].split('\n')[:-1]
			return [x.split()[1] for x in contents]

		def load_package_info(self, package):
			"""From database load informations about given package and set them to it"""
			description = None
			if not package:
				return None

			category = package.name.split('/')[0]
			process = subprocess.Popen(['eix', '-e', package.name], stdout=subprocess.PIPE)
			out = process.communicate()[0]
			out = out.decode().split('\n')

			for line in out:
				line = line.strip()
				if line.startswith("Description:"):
					description = line.split("Description:")[1].strip()

			package.description = description
			package.category = category

		def provided_by(self, app):
			"""Returns a package which provides given application"""
			process = app.instances[0]  # @TODO Reimplement for all processes
			package = self._file_provided_by(process.exe)
			if package:
				if package.category == 'dev-lang':
					for arg in process.cmdline[1:]:
						if os.path.isfile(arg):
							package = self._file_provided_by(arg)
							return package if package else None
				return package
			return None

		def _file_provided_by(self, file):
			find_owner = FileOwner(early_out=True)
			packages = find_owner([file])
			if packages:
				package = packages[0][0]
				name = package.category + '/' + package.name
				p = Package(name)
				p.category = package.category
				return p
			return None



from __future__ import print_function
from __future__ import unicode_literals

from . import View
import sys


class BlocksView(View):

	"""
	Assign ``blocks`` with this list

	[
		{
			"title": "...",
			"content": "some text",
		},
		...
	]
	"""

	def render(self):
		for index, block in enumerate(self.args.blocks):
			if block["content"]:
				if "title" in block:
					self.print(block["title"])
				self.print(block["content"], end="")

				if self.next(index):
					self.print("")

	def next(self, index):
		"""Return True if there is any nonempty block after given index"""
		for block in self.args.blocks[index+1:]:
			if block["content"]:
				return True
		return False

	def has_content_and_title(self):
                """Return True if at least one block's content & title is not empty"""
                for block in self.args.blocks[0:]:
                       if block["content"] and "title" in block.keys():
                                if block["title"]:
                                        return True
                return False



#-*- coding: utf-8 -*-
# rules.py
# Manager for rules file
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

from xml.dom import minidom
from xml.parsers.expat import ExpatError
from tracer.paths import DATA_DIR, USER_CONFIG_DIRS
from tracer.resources.exceptions import PathNotFound, TracerError
from os.path import dirname


class Rules(object):

	DEFINITIONS = map(lambda x: x + "/rules.xml", [DATA_DIR] + USER_CONFIG_DIRS)

	ACTIONS = {
		"CALL-PARENT"  :  "call-parent",
		"RETURN"        :  "return",
	}
	_DEFAULT_ACTION = ACTIONS["CALL-PARENT"]
	_rules = None

	@staticmethod
	def find(app_name):
		if not Rules._rules:
			Rules._load_definitions()

		for rule in Rules._rules:
			if rule.name == app_name:
				return rule

	@staticmethod
	def all():
		if not Rules._rules:
			Rules._load_definitions()

		return Rules._rules

	@staticmethod
	def _load_definitions():
		Rules._rules = []
		for file in Rules.DEFINITIONS:
			try: Rules._load(file);
			except PathNotFound as ex:
				if not dirname(file) in USER_CONFIG_DIRS:
					raise ex

	@staticmethod
	def _load(file):
		try:
			with open(file, "r") as f:
				xmldoc = minidom.parseString(f.read())
		except IOError:
			raise PathNotFound('DATA_DIR')
		except ExpatError as ex:
			msg = "Unable to parse {0}\nHint: {1}".format(file, ex)
			raise TracerError(msg)

		for rules in xmldoc.getElementsByTagName("rules"):
			for rule in rules.getElementsByTagName("rule"):
				attrs = dict(rule.attributes.items())
				r = Rule(attrs)
				if r in Rules._rules:
					i = Rules._rules.index(r)
					Rules._rules[i].update(r)
				else:
					r.setdefault('action', Rules._DEFAULT_ACTION)
					Rules._rules.append(r)


class Rule(object):

	"""
	Represent the rule defined in `rules.xml`

	Attributes
	----------
	name : str
	action : str
		See `Rules.ACTIONS` for possible values
	"""

	_attributes = None

	def __init__(self, attributes_dict):
		self._attributes = attributes_dict

	def __eq__(self, other):
		return isinstance(other, Rule) and self.name == other.name

	def __ne__(self, other):
		return not self.__eq__(other)

	def __getattr__(self, item):
		return self._attributes[item]

	def __len__(self):
		return len(self._attributes)

	def __contains__(self, item):
		return item in self._attributes

	def __str__(self):
		return "<Rule: " + self._attributes["name"] + ">"

	def __repr__(self):
		return self.__str__()

	def setdefault(self, key, value):
		self._attributes.setdefault(key, value)

	def update(self, values):
		if isinstance(values, Rule):
			values = values._attributes
		self._attributes.update(values)



# SystemdUnit.py
# Module for getting data from Systemd about Units

# Copyright (C) 2017 Sean O'Keeffe
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

import dbus

class SystemdDbus(object):
	def __init__(self):
		self.__systemd = dbus.SystemBus().get_object('org.freedesktop.systemd1','/org/freedesktop/systemd1')
		self.__manager = dbus.Interface(self.__systemd, dbus_interface='org.freedesktop.systemd1.Manager')

	def unit_path_from_pid(self, pid):
		try:
			return self.__manager.GetUnitByPID(pid)
		except dbus.exceptions.DBusException:
			return False

	def unit_path_from_id(self, Id):
		try:
			return self.__manager.GetUnit(Id)
		except dbus.exceptions.DBusException:
			return False

	def has_service_property_from_pid(self, pid, attr):
		try:
			unit = self.unit_path_from_pid(pid)
			if not unit:
				return False

			proxy = dbus.SystemBus().get_object('org.freedesktop.systemd1', unit)
			propty = proxy.Get('org.freedesktop.systemd1.Service', attr, dbus_interface='org.freedesktop.DBus.Properties')
		except dbus.exceptions.DBusException:
			return False
		return bool(propty)

	def get_unit_property_from_pid(self, pid, attr):
		unit_path = self.unit_path_from_pid(pid)
		if bool(unit_path):
			proxy = dbus.SystemBus().get_object('org.freedesktop.systemd1', self.unit_path_from_pid(pid))
			return proxy.Get('org.freedesktop.systemd1.Unit', attr, dbus_interface='org.freedesktop.DBus.Properties')
		else:
			return False



#-*- coding: utf-8 -*-
# memory.py
# Module to work with files in memory
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import

from tracer.resources.processes import Processes
import psutil


def dump_memory(user=None):
	"""
	Returns memory in BTree structure

	{ file_1 : [process_1, process_2, ..., process_n],
		...
	}

	Which describes that processes 1 to `n` is using file_1
	"""
	memory = {}
	for process in Processes.all().owned_by(user).unique():
		try:
			for file in process.files:
				if file in memory:
					memory[file].append(process)
				else:
					memory[file] = [process]

		except psutil.NoSuchProcess: pass
		except psutil.AccessDenied: pass

	return memory



#-*- coding: utf-8 -*-
# pycomp.py
# Compatibility layer between python2 and python3
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from sys import version_info

PY3 = version_info.major >= 3

if PY3:
	from io import StringIO
	from functools import lru_cache
else:
	from StringIO import StringIO
	from backports.functools_lru_cache import lru_cache


def load_source(module_name, path):
	"""
	Read and evaluate a python file
	This is useful when we don't know the module name beforehand and somehow
	figure it out at the runtime (e.g. user-defined hook files)
	"""
	if not PY3:
		import imp
		return imp.load_source(module_name, path)

	from importlib.machinery import SourceFileLoader
	import types
	loader = SourceFileLoader(module_name, path)
	loaded = types.ModuleType(loader.name)
	return loader.exec_module(loaded)



#-*- coding: utf-8 -*-
# exceptions.py
# Tracer exceptions module
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from __future__ import absolute_import
from __future__ import print_function

from sys import version_info
from tracer.version import __version__
from tracer.resources.lang import _


class Printable(object):
	message = None
	def print(self):
		print(self.message.encode("utf-8") if version_info.major == 2 else self.message)


class TracerError(Exception, Printable):
	"""
	Unspecified tracer error
	"""
	def __init__(self, message):
		self.message = message


class UnsupportedDistribution(OSError, Printable):

	@property
	def message(self):
		return _(
			"You are running unsupported linux distribution\n"
			"\n"
			"Please visit https://github.com/FrostyX/tracer/issues\n"
			"and create new issue called 'Unknown or unsupported linux distribution: {0} (v{1})' if there isn't such.\n"
			"\n"
			"Don't you have an GitHub account? Please report this issue on frostyx@email.cz")

	def __init__(self, distro):
		OSError.__init__(self, self.message.format(distro, __version__))


class LockedDatabase(OSError, Printable):

	@property
	def message(self):
		return _("Package database is locked by another process")

	def __init__(self):
		OSError.__init__(self, self.message)


class DatabasePermissions(OSError, Printable):

	@property
	def message(self):
		return _("You can't open package database due to insufficient permissions")

	def __init__(self):
		OSError.__init__(self, self.message)


class PathNotFound(OSError, Printable):

	@property
	def message(self):
		return _(
			"Problem occurred - neither one of {0} paths exists\n"
			"Please contact maintainer of tracer package in your distribution.")

	def __init__(self, name):
		OSError.__init__(self, self.message.format(name))



#-*- coding: utf-8 -*-
# version.py
# Version module generated with tito
# WARNING: !! Do not edit this file !!
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

__version__ = '1.4'
__release__ = '1'



#-*- coding: utf-8 -*-
# paths.py
# Module for defining paths to project directories. They are different when
# project is developed on git and when project is installed as a linux package
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

import os
from os.path import dirname, realpath, expanduser
from tracer.resources.system import System


def __(paths):
	for path in paths:
		if os.path.isdir(path):
			return path
	return paths[0]


PROJECT_DIR = dirname(dirname(realpath(__file__)))

DATA_DIR = __([
	PROJECT_DIR + '/' + 'data',
	'/usr/share/tracer',
])

USER_CONFIG_DIRS = [
	'/etc/tracer',
]

HOOKS_DIRS = [
	'/etc/tracer/hooks',
]

LANG_DIR = __([
	PROJECT_DIR + '/build/' + 'locale',
	'/usr/share/locale',
])

try:
	user = System.user()
	USER_CONFIG_DIRS.append(expanduser('~' + user) + '/.config/tracer')
	HOOKS_DIRS.append(expanduser('~' + user) + '/.config/tracer/hooks')
except OSError:
	pass



#-*- coding: utf-8 -*-
# processes.py
# Module providing informations about processes
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#

from .collections import ProcessesCollection
from .FilenameCleaner import FilenameCleaner
import psutil
import datetime
import time
import os
from subprocess import PIPE, Popen
from threading import Timer
from six import with_metaclass


class Processes(object):

	# psutil 3.x to 1.x backward compatibility
	@staticmethod
	def pids():
		try:
			return psutil.pids()
		except AttributeError:
			return psutil.get_pid_list()

	@staticmethod
	def all():
		processes = ProcessesCollection()
		for pid in Processes.pids():
			try:
				processes.append(Process(pid))
			except psutil.NoSuchProcess: pass
			except psutil.AccessDenied: pass
		return processes


class ProcessWrapper(object):
	"""
	Wrapper for ``psutil.Process class``
	Library ``psutil`` is not backward compatible from version 2.x.x to 1.x.x.

	Purpose of this class is cover incompatibility in ``psutil.Process`` class and
	provide interface of new version. It allows using new interface even with
	old version of ``psutil``.

	Note that, for performance reasons, process information is cached at
	object creation. To force a refresh, invoke the ``rebuild_cache()``
	method.
	"""

	def __init__(self, pid=None):
		self._process = psutil.Process(pid)
		self.rebuild_cache()

	def __nonzero__(self):
		return bool(self._process)

	def rebuild_cache(self):
		self._procdict = self._process.as_dict(attrs=['name', 'exe', 'cmdline', 'ppid', 'username', 'create_time'])

	def name(self):
		# Special case for sshd, if its cmd contains the execuatable is must be the daemon
		# else must be the session.
		try:
			if self._attr("name") == 'sshd':
				if self._attr("exe") not in self._attr("cmdline") and len(self._attr("cmdline")) > 1:
					username= self._attr("cmdline")[1].split("@")[0]
					return 'ssh-{0}-session'.format(username)
		except psutil.AccessDenied:
			pass
		return self._attr("name")

	def exe(self):
		return self._attr("exe")

	def cmdline(self):
		return self._attr("cmdline")

	def ppid(self):
		return self._attr("ppid")

	def parent(self):
		return self._attr("parent")

	def username(self):
		return self._attr("username")

	def create_time(self):
		return self._attr("create_time")

	def children(self, recursive=False):
		key = 'children-{0}'.format(recursive)
		if key not in self._procdict:
			try:
				self._procdict[key] = self._process.children(recursive)
			except AttributeError:
				self._procdict[key] = self._process.get_children(recursive)
		return self._procdict[key]

	def _attr(self, name):
		if name not in self._procdict:
			attr = getattr(self._process, name)
			try:
				self._procdict[name] = attr()
			except TypeError:
				self._procdict[name] = attr
		return self._procdict[name]

	def __getattr__(self, item):
		return getattr(self._process, item)

	# psutil 3.x to 1.x backward compatibility
	def memory_maps(self, grouped=True):
		key = 'memory_maps-{0}'.format(grouped)
		if key not in self._procdict:
			try:
				self._procdict[key] = self._process.memory_maps(grouped=grouped)
			except AttributeError:
				self._procdict[key] = self._process.get_memory_maps(grouped=grouped)
		return self._procdict[key]


class ProcessMeta(type):
	"""
	Caching metaclass that ensures that only one ``Process`` object is ever
	instantiated for any given PID. The cache can be cleared by calling
	``Process.reset_cache()``.

	Based on https://stackoverflow.com/a/33458129
	"""

	def __init__(cls, name, bases, attributes):
		super(ProcessMeta, cls).__init__(name, bases, attributes)
		def reset_cache():
			cls._cache = {}
		reset_cache()
		setattr(cls, 'reset_cache', reset_cache)

	def __call__(cls, *args, **kwargs):
		pid = args[0]
		if pid not in cls._cache:
			self = cls.__new__(cls, *args, **kwargs)
			cls.__init__(self, *args, **kwargs)
			cls._cache[pid] = self
		return cls._cache[pid]


class Process(with_metaclass(ProcessMeta, ProcessWrapper)):
	"""
	Represent the process instance uniquely identifiable through PID

	For all class properties and methods, please see
	http://pythonhosted.org/psutil/#process-class

	Below listed are only reimplemented ones.

	For performance reasons, instances are cached based on PID, and
	multiple instantiations of a ``Process`` object with the same PID will
	return the same object. To clear the cache, invoke
	``Process.reset_cache()``. Additionally, as with ``ProcessWrapper``,
	process information is cached at object creation. To force a refresh,
	invoke the ``rebuild_cache()`` method on the object.
	"""

	def __eq__(self, process):
		"""For our purposes, two processes are equal when they have same name"""
		return self.pid == process.pid

	def __ne__(self, process):
		return not self.__eq__(process)

	def __hash__(self):
		return hash(self.pid)

	@staticmethod
	def safe_isfile(file_path, timeout=0.5):
		"""
		Process arguments could be referring to files on remote filesystems and
		os.path.isfile will hang forever if the shared FS is offline.
		Instead, use a subprocess that we can time out if we can't reach some file.
		"""
		process = Popen(['test', '-f', file_path], stdout=PIPE, stderr=PIPE)
		timer = Timer(timeout, process.kill)
		try:
			timer.start()
			process.communicate()
			return process.returncode == 0
		finally:
			timer.cancel()

	@property
	def files(self):
		files = []

		# Files from memory maps
		for mmap in self.memory_maps():
			files.append(FilenameCleaner.strip(mmap.path))

		# Process arguments
		for arg in self.cmdline()[1:]:
			if not os.path.isabs(arg):
				continue

			if Process.safe_isfile(arg):
				files.append(arg)

		return sorted(files)

	def parent(self):
		"""The parent process casted from ``psutil.Process`` to tracer ``Process``"""
		if self.ppid():
			return Process(self.ppid())
		return None

	def username(self):
		"""The user who owns the process. If user was deleted in the meantime,
		``None`` is returned instead."""

		# User who run the process can be deleted
		try:
			return super(Process, self).username()
		except KeyError:
			return None

	def children(self, recursive=False):
		"""The collection of process's children. Each of them casted from ``psutil.Process``
		to tracer ``Process``."""
		children = super(Process, self).children(recursive)
		return ProcessesCollection([Process(child.pid) for child in children])

	@property
	def exe(self):
		"""The absolute path to process executable. Cleaned from arbitrary strings
		which appears on the end."""

		# On Gentoo, there is #new after some files in lsof
		# i.e. /usr/bin/gvim#new (deleted)
		exe = super(Process, self).exe()
		if exe.endswith('#new'):
			exe = exe[0:-4]

		# On Fedora, there is something like ;541350b3 after some files in lsof
		if ';' in exe:
			exe = exe[0:exe.index(';')]

		return exe

	@property
	def is_interpreted(self):
		# @TODO implement better detection of interpreted processes
		return self.name() in ["python"]

	@property
	def is_session(self):
		terminal = self.terminal()
		if terminal is None:
			return None
		parent = self.parent()
		if parent is None or terminal != parent.terminal():
			return True

	@property
	def real_name(self):
		if self.is_interpreted:
			for arg in self.cmdline()[1:]:
				if os.path.isfile(arg):
					return os.path.basename(arg)
		return self.name()

	@property
	def str_started_ago(self):
		"""
		The time of how long process is running. Returned as string
		in format ``XX unit`` where unit is one of
		``days`` | ``hours`` | ``minutes`` | ``seconds``
		"""

		now = datetime.datetime.fromtimestamp(time.time())
		started = datetime.datetime.fromtimestamp(self.create_time())
		started = now - started

		started_str = ""
		if started.days > 0:
			started_str = str(started.days) + " days"
		elif started.seconds >= 60 * 60:
			started_str = str(int(started.seconds / (60 * 60))) + " hours"
		elif started.seconds >= 60:
			started_str = str(int(started.seconds / 60)) + " minutes"
		elif started.seconds >= 0:
			started_str = str(int(started.seconds)) + " seconds"

		return started_str


class AffectedProcess(Process):
	packages = None
	files = None

	def __init__(self, pid=None):
		Process.__init__(self, pid)
		self.packages = set()
		self.files = set()

	def update(self, process):
		self.files = self.files.union(process.files)
		self.packages = self.packages.union(process.packages)



from __future__ import print_function

from . import View
from tracer.resources.lang import _


class NoteForHiddenView(View):
	def render(self):

		if not self.args.args.quiet and (self.args.session_count > 0 or self.args.static_count > 0):

			if self.args.session_count + self.args.static_count == self.args.total_count:
				self.print(_("There are:"))
			else:
				self.print(_("Additionally, there are:"))

			if self.args.session_count > 0:
				self.print("  - " + \
					_("{0} processes requiring restart of your session (i.e. Logging out & Logging in again)")\
						.format(self.args.session_count)
				)

			if self.args.static_count > 0:
				self.print("  - " + _("{0} processes requiring reboot").format(self.args.static_count))




#-*- coding: utf-8 -*-
# FilenameCleaner.py
# Module to stripping the version numers from filenames
#
# i.e.:
#     /lib/libdl-2.19.so   -->   /lib/libdl.so
#     /lib/libncurses.so.5.9   -->   /lib/libncurses.so
#
# This implementation really sux and surely there is a much better
# way how to implement following methods. They are properly tested
# in `test_FilenameCleaner.py` so feel free to refactor them.
#
# Copyright (C) 2016 Jakub Kadlcik
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#


class FilenameCleaner(object):

	@staticmethod
	def strip(filename):
		filename = FilenameCleaner._strip_abnormalities(filename)
		try:
			slash = filename.rindex('/')
			dirname = filename[:slash]
			basename = filename[slash+1:]

			if "." in basename:
				basename = FilenameCleaner._strip_around_so_delimited_dots(basename)
				basename = FilenameCleaner._strip_after_dash(basename)
			return dirname + '/' + basename

		except IndexError: pass
		except ValueError: pass
		return filename

	@staticmethod
	def _strip_abnormalities(filename):
		# Doesnt matter what is after space cause filename ends with first space
		try: filename = filename[:filename.index(' ')]
		except ValueError: pass

		# On Gentoo, there is #new after some files in lsof
		# i.e. /usr/bin/gvim#new (deleted)
		if filename.endswith('#new'):
			filename = filename[0:-4]

		# On Fedora, there is something like ;541350b3 after some files in lsof
		# See issue #9
		if ';' in filename:
			filename = filename[0:filename.index(';')]

		# On Fedora, there is something like .#prelink#.N3n7Rk (deleted) after some files in lsof
		# See issue #9
		if '.#prelink#.' in filename:
			filename = filename[0:filename.rindex('.#prelink#.')]

		return filename

	@staticmethod
	def _strip_around_so_delimited_dots(basename):
		try:
			split = basename.split(".so")
			basename = split[0]
			if len(split) > 1:
				basename += ".so"
			first_dot_i = basename.index(".")
			last_dot_i = basename.rindex(".")
			basename = basename[:first_dot_i] + basename[last_dot_i:]

		except IndexError: pass
		except ValueError: pass
		return basename

	@staticmethod
	def _strip_after_dash(basename):
		try:
			dash = basename.rindex("-")
			dot = basename.index(".", dash)
			if FilenameCleaner._is_version(basename[dash+1:dot]):
				basename = basename[:dash] + basename[dot:]

		except IndexError: pass
		except ValueError: pass
		return basename

	@staticmethod
	def _is_version(string):
		"""
		Returns True if string contains only digits and dots
		"""
		for char in string:
			if (not char.isdigit()) or char == ".":
				return False
		return True



from __future__ import print_function
from __future__ import unicode_literals

from . import View
from tracer.resources.lang import _
from tracer.resources.applications import Applications
from tracer.views.note_for_hidden import NoteForHiddenView
from tracer.views.blocks import BlocksView
from tracer.resources.pycomp import StringIO
from sys import version_info
import re


class DefaultView(View):
	def render(self):

		def with_helpers_content():
			content = ""
			types = [Applications.TYPES["SESSION"], Applications.TYPES["STATIC"], Applications.TYPES["ERASED"]]
			applications = (self.args.applications
							.with_helpers()
							.exclude_types(types)
							.unique()
							.sorted("helper"))
			for application in applications:
				helpers = "; ".join(application.helpers)
				if application.helper_contains_formating and not application.helper_contains_name:
					helpers += "  # {}".format(application.name)
				content += "      " + helpers + "\n"
			return content

		def without_helpers_content():
			content = ""
			apps = (self.args.applications
					.exclude_types(Applications.TYPES["ERASED"])
					.unique()
					.without_helpers()
					.sorted("name"))
			for application in apps:
				content += "      " + application.name + "\n"
			return content

		def erased_content():
			content = ""
			apps = (self.args.applications
					.filter_types([Applications.TYPES["ERASED"]])
					.unique()
					.sorted("name"))
			for application in apps:
				content += "      " + application.name + "\n"
			return content

		def unrestartable_content(app_type):
			content = ""
			applications = (self.args.applications
							.with_helpers()
							.filter_types([app_type])
							.unique()
							.sorted("name"))
			for application in applications:
				content += "      " + application.name + "\n"
			return content

		def note_content():
			content = StringIO()
			view = NoteForHiddenView(content)
			view.assign("args", self.args.args)
			view.assign("total_count", len(self.args.applications))
			view.assign("session_count", self.args.applications.count_type(Applications.TYPES["SESSION"]))
			view.assign("static_count", self.args.applications.count_type(Applications.TYPES["STATIC"]))
			view.render()
			return content.getvalue() if version_info.major >= 3 else content.getvalue().decode("utf8")

		blocks = [
			{"title": "  * " + _("Some applications using:"), "content": with_helpers_content()},
			{"title": "  * " + _("These applications manually:"), "content": without_helpers_content()},
			{"title": "  * " + _("Uninstalled applications:"), "content": erased_content()},
		]

		if self.args.args.all:
			blocks.append({
				"title": "  * " + _("These applications restarting your session:"),
				"content": unrestartable_content(Applications.TYPES["SESSION"])
			})
			blocks.append({
				"title": "  * " + _("These applications rebooting your computer:"),
				"content": unrestartable_content(Applications.TYPES["STATIC"])
			})
		else:
			blocks.append({"content": note_content()})

		view = BlocksView(self.out)
		view.assign("blocks", blocks)
		if view.has_content_and_title():
			self.print(_("You should restart:"))
		view.render()
